#!/usr/bin/env ruby

require "serialport"
require 'net/http'
require 'json'
require './util'

#
# Keep an updated buffer of anchor configurations from cloud
# Provide a serial API towards ble-gateway to support update requests
# Install:
# Any recent ruby version should work e.g. 2.6 or later:
#     ruby --version
# Serial port package/gem is a required dependency:
#     sudo gem install serialport
#

class AnchorConf

  KIND_MASTER = 1
  KIND_SLAVE  = 2
  KIND_TAG    = 3

  def initialize(floorplan_id)
    @floorplan_id = floorplan_id
    @host = "http://minew.ideell.se"
  end

  # Call explicitely to get external database
  def update_db
    puts "update data from cloud ..."
    json = Net::HTTP.get URI "#{@host}/floorplans/#{@floorplan_id}.json"
    return false if json.empty?
    @floorplan = JSON.parse json
    json = Net::HTTP.get URI "#{@host}/sites/#{@floorplan["site_id"]}.json"
    return false if json.empty?
    @site = JSON.parse json
    json = Net::HTTP.get URI "#{@host}/floorplans/#{@floorplan_id}/anchors.json"
    return false if json.empty?
    @anchors = JSON.parse json
    json = Net::HTTP.get URI "#{@host}/customers/#{@site["customer_id"]}/tags.json"
    return false if json.empty?
    @tags = JSON.parse json
    # Create configuration
    return false if @tags.empty?
    @cfg = {}
    @anchors.each do |x|
      h = x.transform_keys(&:to_sym)
      kind = h[:kind].to_i
      if kind == KIND_MASTER || kind == KIND_SLAVE then 
        key = pretty_ble_mac(h[:ble_mac])
        val = { name: h[:name], 
                utime: time_str_to_i(h[:updated_at]),
                kind: h[:kind].to_i,
                master_id: h[:master_id].to_i,
                slave_cnt: h[:cell_size].to_i 
              }
        val[:slave_ix] = h[:slave_ix].to_i if kind == KIND_SLAVE
        @cfg[key] = val
      else
        puts "unsupported kind"
      end
    end
    @tags.each do |x|
      h = x.transform_keys(&:to_sym)
      kind = h[:kind].to_i
      key = pretty_ble_mac(h[:ble_mac])
      val = { name: h[:name], 
              utime: time_str_to_i(h[:updated_at]),
              kind: KIND_TAG,
            }
      @cfg[key] = val
    end
    return true
  end

  # Get all config values as configuration string
  def all_to_str
    self.class.to_str(all)
  end

  # Get config values with :utime greater than gt_utime as configuration string
  def recent_to_str(gt_utime)
    self.class.to_str(recent(gt_utime))
  end

  private     # only helpers below ...
  
  # Get all config items
  def all
    @cfg
  end

  # Get all config items where :utime is grater than gt_utime
  # TODO: Replace by date string instead?
  def recent(gt_utime)
    @cfg.select { |_, x| x[:utime] > gt_utime }
  end

  # Convert MAC formatted as "09:03:2F:2F:9A:37" with colons removed "09032F2F9A37"
  def self.mac_format(s)
    s.split(':').join
  end

  # Convert GFG key value item to config string
  # Sample format for master: "\x37\x9A\x2F\x2F\x03\x09,M,1,5;"
  def self.item_to_str(k, v)
    s = ""
    mac = mac_format(k)
    case v[:kind]
    when KIND_MASTER
      "#{mac},M,#{v[:master_id]},#{v[:slave_cnt]}"
    when KIND_SLAVE
      "#{mac},S,#{v[:master_id]},#{v[:slave_ix]},#{v[:slave_cnt]}"
    when KIND_TAG
      "#{mac},T"
    end
  end

  # Get max utime value of items
  def self.utime_max(items)
    items.values.max { |a, b| a[:utime] <=> b[:utime] }[:utime]
  end

  # Helper: Convert items to string format
  def self.to_str(items)
    ans = ""
    unless items.empty? then
      # Items using semicolon as separator
      ans = items.map { |k, v| item_to_str(k, v) }.join(';') + ";"
      # add max utime for transmission
      ans << "TIME:" + utime_max(items).to_s + ";"
    end
    # add explicit EOT and semicolon as message end
    ans << "EOT;"
  end
  
end

# Debug output
def show(s)
  puts s
end

if ARGV.size != 2 || ARGV[0] == "help" then
  puts "usage:"
  puts "1. Start BLE gateway device connected via USB"
  puts "2. Determine which device is connected"
  puts "    * Look for \"Product: USB CDC For PALOS\" using e.g. dmesg"
  puts "3. Determine which floorplan gateway is associated with"
  puts "    * Dig down into http://minew.ideell.se/customers"
  puts "    * customer -> site -> floorplan"
  puts "4. Start with serial device as first argument and floorplan id as second e.g."
  puts "    ./cloud-link.rb /dev/ttyACM0 1"
  exit
end

# serial_dev="/dev/ttyACM2"                   # typical Linux device
# serial_dev="/dev/cu.usbmodemC9032F2F9A371"  # typical macOS device

serial_dev = ARGV[0]
floorplan_id = ARGV[1].to_i

db = AnchorConf.new(floorplan_id)
sp = nil

loop do

  show "Update configuration from cloud"

  loop do
    if db.update_db then
      show "Updated configuration from cloud:"
      show (db.all_to_str).split(';').join(";\n")
      break
    else 
      show "Failed to get data from server. Wait ..."
      sleep 10
    end
  end
  
  # Test: print database records
  # show db.all_to_str
  
  show "Start protocol on device: " + serial_dev
  
  # Open protocol towards serial client
  loop do
    # For basic testing ...
    # ioreg -p IOUSB    # macOS
    # usbdevs           # OpenBSD and Linux
    dev=`usbdevs | grep "PALOS"`
    if dev.empty? then
      show "BLE device not connected to USB port. Wait ..."
      sleep 10
    else
      show "BLE device connected. Connect to USB serial port."
      sp = SerialPort.new(serial_dev, 115200, 8, 1, SerialPort::NONE)
      break
    end
  end

  buffer = ""
  cmds = []
  loop do
    begin
      # add data to buffer
      data = sp.getc
      if data then
        if data.size > 0 then
          # show "char: #{data}"
          buffer << data
        end
      else
        show "Serial closed. Wait ..."
        sleep 10
        break
      end
    rescue
      show "IO Error (getc). Wait ..."
      sleep 10
      break
    end

    # extract zero or more commands
    while (not buffer.empty?) && buffer.include?(';') do
      cmd, rest = buffer.split(';', 2)
      cmds << cmd
      buffer = rest
    end
    # print/execute commands
    cmds.each do |cmd|
      show "command: #{cmd}"
      token, n = cmd.split(':', 2)
      case token
      when "nop"
        show "response: N/A"
      when "get"
        # print response
        resp = db.recent_to_str(n.to_i)
        show "response: #{resp}"
        sp.write(resp)
        sp.flush()
        # Alternatively slow down if you have problems ...
        # resp.each_char { |c| 
        #   sp.putc(c)
        #   sleep 0.01
        # }
      else
        show "unsupported command: #{token}"
      end
    sleep 1
    end
    cmds = []
  end

end